/*
 * powwow  --  mud client with telnet protocol
 *
 * Initially inspired to the Tintin client by Peter Unold, 
 * Powwow contains no Tintin code.
 * The original program Cancan,  written by Mattias Engdegrd (Yorick)
 * (f91-men@nada.kth.se) 1992-94,
 * was greatly improved upon by Vivriel, Thuzzle and Ilie and then
 * transformed from Cancan into Powwow by Cosmos who worked
 * to make it yet more powerful.
 * AmigaDOS porting attempt by Fror.
 * Many new features added by Dain.
 * As usual, all the developers are in debt to countless users 
 * for suggestions and debugging.
 *
 * This program is placed in the public domain.
 */

#define POWWOW_VERSION "1.0.5 \n(Yorick, Vivriel, Thuzzle, Ilie, Fror, Dain, Cosmos 1992-98)"
#define SAVEFILEVER 5
#define HELPNAME "powwow_help"

/*
 * Set this to whatever you like
 *
 * #ifndef DEFAULT_HELP_DIR
 * #  define DEFAULT_HELP_DIR	"/home/gustav/powwow"
 * #endif
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <memory.h>
#include <unistd.h>

#include "powwow.h"
#include "beam.h"
#include "commands.h"
#include "edit.h"
#include "mapping.h"
#include "list.h"
#include "tcp.h"
#include "tty.h"
#include "eval.h"

#ifdef USE_RANDOM
#  define get_random random
#  define init_random srandom
#else
#  define get_random lrand48
#  define init_random srand48
#endif

/*         function declarations                */
void mainlupe();
void exec_delays();
void printmotd();
void get_user_input();
void putbackcursor();
void set_param();
void restore_param();
int search_action();
int search_prompt();
int printstrlen();
void put_marks();
void wrap_print();
void smart_print();
void intrsig();
void termsig();
void pipesig();
void parse_commands();
void denewbiefy();
void unescape();
void showprompt();
void get_remote_input();
void writetofile();
void put_word();
void default_completions();
static void demote_word();
void set_custom_delimeters();

extern int errno, sys_nerr;
extern int select();
extern int strncasecmp();

/*        GLOBALS                */
char *helpname = HELPNAME;
long stat_var[NUMVAR];

struct {
    long num[NUMPARAM];
    char *str[NUMPARAM];
} def_param;

param_stack paramstk;
vars *var;                   /* pointer to all variables */

long received = 0;           /* amount of data received from remote host */
long sent = 0;               /* amount of data sent to remote host */

char identified = 0;	     /* 1 after #identify */
char can_suspend = 0;        /* 1 if shell has job control */
static char ready = 1;       /* 0 if powwow is parsing something */
static char confirm = 1;     /* 1 if just tried to quit */
char history_done = 0;       /* 1 if last command contained a #history */   
int internal;                /*   1 if internal echo, 
			      *  -1 if something sent to MUD,
			      *  -2 if #emulate used 
			      */
char opt_exit = 0;		/* 1 to autoquit when closing last conn. */
char opt_history;		/* 1 if to save also history */
char opt_words = 0;		/* 1 if to save also word completion list */
char echo_ext = 1;		/* 1 if text sent to MUD must be echoed */
char echo_key = 1;		/* 1 if binds must be echoed */
char echo_int = 1;		/* 1 if internal messages are not suppressed */
char opt_compact = 0;		/* 1 if to clear prompt between remote messages */
char opt_debug = 0;		/* 1 if to echo every line before executing it */
char opt_speedwalk = 0;		/* speedwalk on? */
char opt_wrap = 1;		/* 1 = word wrap active */
char opt_autoprint = 0;		/* 1 = automatically #print lines matched by actions */

char hostname[LINELEN], portnumber[LINELEN];
char powwow_dir[LINELEN];          /* default path to definition files */
char deffile[LINELEN];             /* name and path of definition file */
char helpfile[LINELEN];            /* name and path of help file */
aliasnode *aliases;                /* head of alias list */
actionnode *actions;               /* head of action list */
promptnode *prompts;               /* head of prompt list */
marknode *markers;                 /* head of mark list */
int a_nice = 0;                    /* default priority of new actions/marks */
keynode *keydefs;                  /* head of key binding list */
delaynode *delays;                 /* head of delayed commands list */
delaynode *dead_delays;            /* head of dead-delayed commands list */
varnode *named_vars[2];            /* head of named variables list */
int num_named_vars[2];             /* number of named variables defined */
int max_named_vars = 100;          /* initial max number of named variables
				    * (automatically increased when needed) */

varnode *variable_prompt;	   /* GH: $prompt is always set		  */
varnode *variable_line;            /*     $line is always set to the last
                                    *     line processed		  */

int word_delim_mode = DELIM_NORMAL;  /* GH: normal word delimeters ( ;) */

vtime now;                         /* current time */
vtime start_time;                  /* time of powwow timer startup */
vtime ref_time;                    /* time corresponding to timer == 0 */

void (*last_edit_cmd)() = 0;	   /* GH: keep track of for repeated cmds */

clock_t start_clock, cpu_clock;

int waiting_for_pid;               /* = pid, when executing #! or editor */

char initstr[LINELEN];             /* initial string to send on connect */

int linemode = 0;                  /* line mode flags (LM_* in powwow.h) */

/* for line editing */
int cols=80, lines=24;    /* screen size */
int col0;                 /* input line offset (= length of prompt) */
int line0;                /* screen line where the input line starts */
char edbuf[LINELEN];      /* line editing buffer */
int edlen;                /* length of current input line */
int pos = 0;              /* cursor position in line */
char promptstr[LINELEN];  /* entire prompt string (for redrawing) */
int promptlen = 0;        /* length of prompt string */
char surely_isprompt = 0; /* 1 if last #prompt set #isprompt */
char verbatim = 0;        /* 1 = don't expand aliases or process semicolons */
char prefixstr[LINELEN];  /* inserted in the editing buffer each time */
char inserted_next[LINELEN];/* inserted in buffer just once */
char flashback = 0;       /* cursor is on excursion and should be put back */
int excursion;            /* where the excursion is */
char edattrbeg[CAPLEN];   /* starting input line attributes */
char edattrend[CAPLEN];   /* ending input line attributes */
int need_prompt_redraw = 0;	/* GH: set to true when window size changes */
int window_size_changed = 0;	/* GH: moved this out of sigwinch handler   */

/* history buffer */
char *hist[MAXHIST];    /* saved history lines */
int curline = 0;        /* current history line */
int pickline = 0;        /* line to pick history from */

/* GH: different ID characters for different action types */
/*
 * Cosmos: they are hardcoded in commands.c, line 1256-1258
 * so don't change them.
 */
char action_chars[ACTION_TYPES] = { '>', '%' };


/* GH: different delimeter modes */
char *delimeters[DELIM_MODES] = { DELIM, " <>!=(),.;\"'{}[]+-/*%", 0 };
char *delim_mode_names[DELIM_MODES]={ "normal", "program", "custom" };

/* 
 * GH: completion list, stolen from cancan 2.6.3a
 *	
 *     words[wordindex] is where the next word will go (always empty)
 *     words[wordindex].prev is last (least interesting)
 *     words[wordindex].next is 2nd (first to search for completion)
 */
#define WORD_UNIQUE	1		/* word is unique in list */
#define	WORD_RETAIN	2		/* permanent (#command) */
typedef struct {
    char *word;
    int  next, prev;
    char flags;
} wordnode;

wordnode words[MAXWORDS];
int wordindex = 0;

FILE *capfile = (FILE *)NULL;        /* capture file or NULL */
FILE *recordfile = (FILE *)NULL;     /* record file or NULL */

void main(argc, argv)
int argc; char *argv[];
{
    char *p;
    int i;
    int read_file = 0;		/* GH: if true, powwow was started with a file
				 *     argument, and initstr shall be ran
				 */
    
    /* initializations */
    initstr[0] = 0;
    bzero(conn_list, sizeof(conn_list));
    
    gettimeofday(&start_time, NULL);
    now.tv_sec  = ref_time.tv_sec  = start_time.tv_sec;
    now.tv_usec = ref_time.tv_usec = start_time.tv_usec;
    start_clock = cpu_clock = clock();
#ifndef NO_RANDOM
    init_random((int)now.tv_sec);
#endif
    
    paramstk.curr = 0;      /* stack is empty */
    
    strcpy(helpfile, (p = getenv("POWWOWHELP")) ? p :
#ifdef DEFAULT_HELP_DIR
	   DEFAULT_HELP_DIR
#else
	   "."
#endif
	   );
    
    if (helpfile[strlen(helpfile) - 1] != '/')
      strcat(helpfile, "/");
    strcat(helpfile, helpname);
    if (access(helpfile, R_OK) == -1 && !access(helpname, R_OK))
      strcpy(helpfile, helpname);

    if ((p = getenv("POWWOWDIR")) != NULL) {
	strcpy(powwow_dir, p);
	if (powwow_dir[strlen(powwow_dir) - 1] != '/')
	    strcat(powwow_dir, "/");
    } else
	powwow_dir[0] = '\0';

    if ((var = (vars *)malloc((NUMVAR+NUMPARAM+max_named_vars)*sizeof(vars)))) {
	for (i=0; i<NUMVAR; i++) {
	    var[i].num = &stat_var[i];
	    var[i].str = NULL;
	}
	for (i=NUMVAR; i<NUMVAR+NUMPARAM; i++) {
	    var[i].num = &def_param.num[i];
	    var[i].str = def_param.str[i] = NULL;
	}
    }
    else
      syserr("malloc");

    /* GH: allocate permanent variables; hack warning! */
    variable_prompt = add_varnode("prompt", 1);
    free(variable_prompt->str);
    var[variable_prompt->index + NUMVAR].str = variable_prompt->str = promptstr;
    variable_line = add_varnode("last_line", 1);

    default_completions();

    tty_bootsrap();
    
#ifdef MOTDFILE
    printmotd();
#endif
    
    printver();

    if (argc == 1 || argc == 3) {
	tty_add_initial_binds();
	tty_add_walk_binds();
    } else if (argc == 2 || argc == 4) {
        /*
         * assuming first arg is definition file name
         * If three args, first is definition file name,
         * second and third are hostname and port number
         * (they overwrite the ones in definition file)
         */
        set_deffile(argv[1]);
	
        if (access(deffile,R_OK) == -1 || access(deffile,W_OK) == -1) {
            tty_printf("Creating %s\nHost name  :", deffile);
	    tty_flush();
	    tty_gets(hostname, LINELEN);
	    if (hostname[0] == '\n')
	      hostname[0] = '\0';
	    else
	      strtok(hostname, "\n");
            tty_puts("Port number:");
	    tty_flush();
	    tty_gets(portnumber, LINELEN);
	    if (portnumber[0] == '\n')
	      portnumber[0] = '\0';
	    else
	      strtok(portnumber, "\n");
            tty_add_initial_binds();
	    tty_add_walk_binds();
            if (save_settings() < 0)
	      exit(0);
        } else if (read_settings() < 0)
	    exit(0);
	else
	    read_file = 1;
    } 
    if (argc == 3 || argc == 4) {
        /* assume last two args are hostname and port number */
        strcpy(hostname, argv[argc - 2]);
        strcpy(portnumber, argv[argc - 1]);
    }
    
    tty_init();
    signal_init();
    tty_window_change();
    
    tty_puts(tty_clreoscr);

    mainlupe(read_file && !*hostname);
}

/*
 * set definition file:
 * rules: if powwow_dir is set it is searched first.
 *        if file doesn't exist, it is created there.
 * If a slash appears in the name, the powwow_dir isn't used.
 */
void set_deffile(arg)
char *arg;
{
    if (strchr(arg, '/') == NULL && *powwow_dir) {
	strcpy(deffile, powwow_dir);
	strcat(deffile, arg);
	if ((access(deffile, R_OK) == -1 || access(deffile, W_OK) == -1)
	    && !access(arg,R_OK) && !access(arg, W_OK))
	    strcpy(deffile, arg);
    } else
	strcpy(deffile, arg);
}

/*
 * show current version
 */
void printver()
{
    tty_printf("Powwow version %s\n%s (%s, %s)%s\n", POWWOW_VERSION, 
#if __STDC__
	       "compiled " __TIME__ " " __DATE__,
#else
	       "", 
#endif
#ifdef USE_VT100
	       "vt100-only",
#else
	       "termcaps",
#endif
#ifdef USE_TERMIO
	       "termios",
#else
	       "BSD sgtty",
#endif
#ifdef USE_REGEXP
	       ""
#else
	       " no regexp"
#endif
	       );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if (f) {
        while (fgets(line, LINELEN, f))
	  tty_puts(line);
	fclose(f);
    }
}
#endif

/*
 * GH: memory-"safe" strdup
 */
char *str_dup(s)
char *s;
{
    char *k;
    
    if (!s)
      return s;
    if ((k = (char *)malloc(strlen(s) + 1)))
      strcpy(k, s);
    return k;
}

/*
 * main loop.
 *
 * GH: if run_initstr, initstr will be ran before the loop starts
 */
void mainlupe(run_initstr)
int run_initstr;
{
    fd_set readfds;
    int err;
    struct timeval timeoutbuf, *timeout;
    long sleeptime;
    
    static int olines;
    olines = lines;

    FD_ZERO(&fdset);
    FD_SET(tty_read_fd, &fdset);
    
    if (*hostname && *portnumber) {
	tcp_open("main", (*initstr ? initstr : NULL), hostname, 
		 atoi(portnumber));
	tcp_write_tty_size();
    }
    
    tty_putc('\n');
    tty_gotoxy(col0 = 0, lines - 2);
    
    tty_puts("Type #help for help.\n");
    line0 = lines - 1;
    
    if (run_initstr && initstr)
	parse_instruction(initstr, 0, 0);

    confirm = 0;
    
    /*                main lupe             */
    
    for (;;) {
	readfds = fdset;
	
	gettimeofday(&now, NULL); 
	tcp_fd = tcp_main_fd;
	exec_delays();
	
	tcp_flush(); /* SPLOOOOSH */
	tty_flush();
	
	sleeptime = 0;
	if (delays) {
	    sleeptime = diff_vtime(&delays->when, &now);
	    if (!sleeptime)
	      sleeptime = 1;    /* if sleeptime is less than 1 millisec,
				 * set to 1 millisec */
	}
	if (flashback && (!sleeptime || sleeptime > FLASHDELAY))
	  sleeptime = FLASHDELAY;
	
	if (sleeptime) {
	    timeoutbuf.tv_sec = sleeptime / mSEC_PER_SEC;
	    timeoutbuf.tv_usec = (sleeptime % mSEC_PER_SEC) * uSEC_PER_mSEC;
	    timeout = &timeoutbuf;
	} else
	  timeout = (struct timeval *)NULL;
	
	ready = 1;
	do {
	    err = select(MAXFDSCAN, &readfds, NULL, NULL, timeout);

	    if (window_size_changed) {
		tcp_write_tty_size();
		window_size_changed = 0;
	    }
	    if (need_prompt_redraw) {
		line0 += lines - olines;

		tty_gotoxy(0, line0);
		tty_puts(tty_clreoscr);
		/* so we know where the cursor is */
		tty_printf("%s%s", edattrend, promptstr);
		tty_draw_input_line();

		tty_flush();			/* GH: probably needed */

		olines = lines;
		need_prompt_redraw = 0;
	    }
	} while (err < 0 && errno == EINTR);
	if (err < 0 && errno != EINTR)
	    syserr("select");
	ready = 0;
	
	if (flashback) putbackcursor();
	
	gettimeofday(&now, NULL);
	
	if (FD_ISSET(tty_read_fd, &readfds))  {
	    err--;
	    tcp_fd = tcp_main_fd;
	    get_user_input();
	}
	
	for (tcp_fd = TCP_MIN_FD; err && (tcp_fd < MAXFDSCAN); tcp_fd++)
	  if (FD_ISSET(tcp_fd, &readfds)) {
	      err--;
	      get_remote_input();
	  }
    }
}

/*
 * execute the delayed labels that have expired
 * and place them in the disabled delays list
 */
void exec_delays()
{
    delaynode *dying;
    char buf[LINELEN];
    
    while (delays && cmp_vtime(&delays->when, &now) <= 0) {
	dying = delays;           /* remove delayed command from active list */
	delays = dying->next;     /* and put it in the dead one  */
	
	add_node((defnode *)dying, (defnode **)&dead_delays, rev_time_sort);
	
	/* must be moved before executing delay->command
	 * and command must be copied in a buffer
	 * (can't you imagine why? The command may edit itself...)
	 */
	
	tty_clear_input_line(opt_compact);
	if (!opt_compact) {
	    tty_putc('\n');
	    col0 = 0;
	}
	
	tty_puts(edattrend);
	if (echo_int)
	  tty_printf("#now [%s]\n", dying->command);
	
	error = 0; internal = 0;
	strcpy(buf, dying->command);
	parse_instruction(buf, 0, 0);
	history_done = 0;
	if (internal >= 0) {
	    showprompt();
	    col0 = printstrlen(promptstr);
	}
	else if (internal != -2)
	  promptlen = promptstr[0] = surely_isprompt = '\0'; 
	
	tty_draw_input_line(); /* it also prints edattrbeg */
    }
}


#define ISNEWLINE(s, len) (*(s)=='\n' || (*(s)=='\r' && (len) > 1 && (s)[1]=='\n'))

/*
 * get data from the socket and process/display it.
 */
void get_remote_input()
{
    char realbuffer[BUFSIZE + 2];        /* allow for a terminating \0 later */
    char *buffer, *buf;
    char deleteprompt = 0, mergeprompt = 0, actions_on_prompt = 0;
    int got, otcp_fd;
    
    got = tcp_read(buffer = realbuffer + promptlen, BUFSIZE - promptlen);
    if (!got) return;
    
    buffer[got]='\0';  /* Safe, there is space. Do it now not to forget it later */
    received += got;
    
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    {
	char c, *t;
	buf = buffer + got;
	tty_putc('{');
	for (t = buffer; t < buf; t++) {
	    if ((c = *t) < ' ' || c > '~')
	      tty_printf("[%d]", c);
	    else tty_putc(c);
	}
	tty_puts("}\n");
    }
#endif
    
    if (!CONN_LIST(tcp_fd).active)
	return; /* process only active connections */

    if (CONN_LIST(tcp_fd).mode == 1) {
	/* this is data from a spawned child or an attached program.
	 * execute as if typed */
	otcp_fd = tcp_fd;
	tcp_fd = tcp_main_fd;
	if (buffer[got-1] == '\n')
	    buffer[got-1] = '\0';
	parse_user_input(buffer, 0);
	tcp_fd = otcp_fd;
	return;
    }
			 
    if (capfile && tcp_fd == tcp_main_fd)
	writetofile(capfile, buffer, got);
    
    if (linemode & LM_CHAR) {
	/* char-by-char mode: just display output, no fuss */
	tty_puts(buffer);
	return;
    }

    /* line-at-a-time mode: process input in a number of ways */
    
    buf = buffer;
    
    /* 
     * Code to merge lines from host that were splitted
     * into different packets: it is a horrible kludge (sigh)
     * and can be used only on one connection at time.
     * We currently do it on main connection.
     * 
     * Note that this code also works for _prompts_ splitted into
     * different packets, as long as no #prompts execute #isprompt
     * on an incomplete prompt.
     */
    
    /*
     * Don't merge if the first part of the line was intercepted
     * and recognized as a prompt by a #prompt action
     * (to avoid intercepting it twice)
     */
    if (tcp_fd == tcp_main_fd && !surely_isprompt && promptstr[0]) {
	if (ISNEWLINE(buf,got)) {
	    /*
	     * the prompt was actually a complete line.
	     * no need to put it on the top of received data,
	     * just match #actions on it
	     */
	    actions_on_prompt = 1;
	} else {
	    /*
	     * put the first part of the splitted line
	     * (currently in the prompt buffer)
	     * on the top of buffer, pretending we have just received it,
	     * then clear the prompt buffer.
	     */
	    bcopy(promptstr, realbuffer, promptlen);
	    buf = buffer = realbuffer;
	    got += promptlen;
	    mergeprompt = 1;
	}
    }

    if (tcp_fd == tcp_main_fd && opt_compact && ISNEWLINE(buf, got)) {
	/* in compact mode, skip the first \n */
	if (*buf == '\n') {
	    buf++; got--;
	    if (got && *buf == '\r') {
		buf++; got--;    /* some MUDs like to send "\n\r" ... */
	    }
	}
	if (*buf == '\r') {
	    buf++; got--;
	    if (got && *buf == '\n') {
		buf++; got--;    /* ... or "\r\n" */
	    }
	}
	deleteprompt = 1;
    }
    
    if (!got && !actions_on_prompt) {
      if (deleteprompt) {
	  tty_puts(edattrend);
	  tty_clear_input_line(deleteprompt);
      }
      return;
    }
        
    tty_puts(edattrend);
    
    if (mergeprompt) {
	tty_clear_input_line(1);
	promptstr[0] = promptlen = 0;
    } else if (actions_on_prompt) {
	/*
	 * kludge for kludge. don't delete the old prompt,
	 * don't print again the line if it doesn't match,
	 * then delete it if opt_compact != 0.
	 */
	tty_clear_input_line(0);
	tty_puts(tty_begoln);
	col0 = promptlen = 0;
	strncpy(variable_line->str, promptstr, PARAMLEN);
	variable_line->str[PARAMLEN-1] = '\0';
	
	if (search_action(promptstr) || opt_compact) {
	    if (line0 < lines - 1)
		tty_puts(tty_clreoscr);
	    else
		tty_puts(tty_clreoln);
	}
	promptstr[0] = '\0';
    } else
	tty_clear_input_line(deleteprompt);
    
    process_remote_input(buf, got, deleteprompt);  /* GH: added last param */
    
    /*        tty_puts(edattrbeg);   */
    /*  No need to print it, tty_draw_input_line() does it itself */
    
    tty_draw_input_line();
}

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to tty
 */
void process_remote_input(buf, size, prompt_deleted)
char *buf; int size; int prompt_deleted;
{
    char onlyprompt = !prompt_deleted;            /* GH: changed from = 1 */
    char *linestart, *lineend, *end = buf + size;
    int mpilen = strlen(MPI);
    char buf2[LINELEN];
    int found_nl;		/* GH: added to allow \n-ending */
    static int last_was_nl = 0; /* Cosmos: fix MPI-in-whois security hole */
    
    if (tcp_fd == tcp_main_fd)
	*promptstr = promptlen = 0;
    
    do {
	/*
	 * process one line at time. stop at "\n" or "\r\n"
	 * also remove extra "\r" before or after "\n"
	 */
        found_nl = 0;                /* GH: added to allow \n-ending */

        linestart = lineend = buf;
	for (; buf < end; buf++) {
	    if (*buf == '\n') {
		/* got a newline: \n */
		last_was_nl = 1;
		lineend = buf;
		*buf = '\0';
		if (buf > linestart && buf[-1] == '\r') {
		    /* well, it was \r\n . just remove the \r */
		    buf[-1] = '\0';
		    lineend--;
		}
		buf++;
		if (buf < end && *buf == '\r') {
		    /*
		     * this was \n\r instead. remove the \r as well,
		     * and also forbid any MPI (MUME kludge)
		     */
		    *buf++ = '\0';
		    last_was_nl = 0;
		}
		found_nl = 1;
		break;
	    }
	}

	if (buf == end && !found_nl)
	  lineend = end;  /* line continues till end of buffer, no trailing \n */

	if (last_was_nl && !strncmp(MPI, linestart, mpilen)) {
	    if (!identified) {
	        PRINT_INT("#warning: MPI message received without #identify\n");
	    } else {
	        /* got a MUME MPI message, don't send it to screen! */
	        buf = lineend = linestart + mpilen +
		  process_message(linestart + mpilen, end - linestart - mpilen);
		continue;
	    }
	}

#ifdef DEBUGCODE_2
	/* debug code to see in detail what codes come from the server */
	{
	    char c;
	    char *t;
	    tty_putc('{');
	    for (t = linestart; t < buf && (c = *t); t++) {
		if (c < ' ' || c > '~')
		  tty_printf("[%d]", c);
		else tty_putc(c);
	    }
	    tty_puts("}\n");
	}
#endif

	/*
	 * try to guess where is the prompt... really not much more than
	 * a guess. again, do it only on main connection: we do not want
	 * output from other connections to mess with the prompt
	 * of main connection :)
	 * 
	 * Since we now have #prompt, behave more restrictively:
	 * if a line has no final \n and is short enough,
	 * we check for #prompts on it.
	 * if no #prompts match or a #prompt matches but does not set #isprompt
	 *     (i.e. recognize it for not being a prompt),
	 *     we check for #actions on it when the \n arrives.
	 * if a #prompt matches and sets #isprompt, then it is REALLY a prompt
	 *     so never match #actions on it.
	 */
	if (tcp_fd == tcp_main_fd && lineend == end && printstrlen(linestart) < cols) {
 	    if (onlyprompt && !strchr(linestart, '\b'))
	      strcat(promptstr, linestart);
	    else
	      strcpy(promptstr, linestart);    /* MUME kludge */

	    if (*promptstr) {
		
		/*
		 * Cosmos: do not match #actions on a prompt.
		 * Match #prompts instead :)
		 * Anyway, remember they could be the first part of a normal line
		 * and the user can have an #action for it.
		 */
		if (!surely_isprompt)
		    search_prompt(promptstr);
		
		/*
		 * now draw the prompt. recalculate length, as search_prompt()
		 * might have changed it.
		 */
		col0 = printstrlen(promptstr);
		promptlen = strlen(promptstr);
	    		
		put_marks(buf2, promptstr);
		tty_puts(buf2);
	    }
	} else {
	    onlyprompt = surely_isprompt = 0;
	    if (linestart[0])
	        strcpy(variable_line->str, linestart);
	    /* GH: only search non-null lines */
	    if ((linestart[0] && !search_prompt(linestart) && !surely_isprompt
		&& (!search_action(linestart) || opt_autoprint))
		|| !linestart[0]) {
		
		if (line0 < lines - 1) line0++;
		if (tcp_fd != tcp_main_fd)        /* sub connection */
		  tty_printf("##%s> ", CONN_LIST(tcp_fd).id);
		
		smart_print(linestart);
	    }
	    col0 = 0;
	}
    } while (lineend < end);
}

/*
 * Determine the printed length of a string. This is sometimes less than
 * the string length since it might contain escape sequences. Treated as such
 * are "<esc> [ <non-letters> <letter>", "<esc> <non-[>", "<control-char>".
 * This is not entirely universal but covers the most common cases (i.e. ansi)
 */
int printstrlen(s)
char *s;
{
    int l;
    enum { NORMAL, ESCAPE, BRACKET } state = NORMAL;
    for (l = 0; *s; s++) {
        switch(state) {
	 case NORMAL:
            if (*s == '\033') {
                state = ESCAPE;
            } else if (*s >= ' ') {
                l++;
            }
            break;
	    
	 case ESCAPE:
            state = (*s == '[') ? BRACKET : NORMAL;
            break;
	    
	 case BRACKET:
            if (isalpha(*s))
	      state = NORMAL;
            break;
        }
    }
    return l;
}

#ifdef USE_REGEXP

/*
 * GH: matches precompiled regexp, return actual params in param array
 *     return 1 if matched, 0 if not
 */
int match_regexp_action (regexp, line)
regex_t *regexp; char *line;
{
    regmatch_t match[NUMPARAM - 1];

    if (!regexec(regexp, line, NUMPARAM - 1, match, 0)) {
	int n, len;

	for (n = NUMVAR; n < NUMPARAM + NUMVAR; n++)
	    *var[n].str = *var[n].num = 0;

	strcpy(var[NUMVAR].str, line);
	n = 0;
	while (n <= regexp->re_nsub && match[n].rm_so != -1 &&
	       n < NUMPARAM - 1) {
	    strncpy(var[NUMVAR + 1 + n].str, line + match[n].rm_so, 
		    len = match[n].rm_eo - match[n].rm_so);
	    var[NUMVAR + ++n].str[len] = 0;
	}
	return 1;
    }
    return 0;
}

#endif

/*
 * match action containing &1..&9 and $1..$9 and return actual params in
 * param array - return 1 if matched, 0 if not
 */
int match_weak_action (pat, src)
char *pat, *src;
{
    char mpat[LINELEN], *npat=0, *npat2=0, *nsrc=0, *prm=0, c;
    char *tmp, *line=src, *realpat = pat;
    int mbeg = 0, mword = 0, p;
    
    for (p = 0; p < NUMPARAM; p++) {
	*var[p+NUMVAR].str = '\0';
	*var[p+NUMVAR].num = 0;
    }
    
    if (*pat == '^') {
	pat++;
	mbeg = 1;  /* anchor match at line start */
    }
    if (*pat == '&' || *pat == '$')
      mbeg = - mbeg - 1;  /* pattern starts with '&' or '$' */
    
    while (pat && *pat) {
	if (((c=*pat) == '&' || c == '$')) {  /* &x matches a string */
	    /* $x matches a single word */
	    tmp = pat + 1;
	    if (isdigit(*tmp)) {
		p = 0;
		while (isdigit(*tmp) && p < NUMPARAM) {
		    p *= 10;
		    p += *tmp++ - '0';
		}
		if (p <= 0 || p >= NUMPARAM)
		  return 0;
		prm = var[p + NUMVAR].str;
		pat = tmp;
		if (c == '$')
		  mword = 1;
	    } else {
		PRINT_INT ("#error: bad action pattern `%s\'\n#missing digit after `%c\'\n",
			   realpat, *pat);
		return 0;
	    }
	}
	
	npat  = first_valid(pat, '&');
	npat2 = first_valid(pat, '$');
	if (npat2 < npat) npat = npat2;
	if (!*npat) npat = 0;
	
	if (npat) {
	    strncpy(mpat, pat, npat-pat);
	    mpat[npat - pat] = 0;
	}
	else
	  strcpy(mpat, pat);
	
	if (*mpat) {
	    nsrc = strstr(src, mpat);
	    if (!nsrc) return 0;
	    if (mbeg > 0) {
		if (nsrc != src) return 0;
		mbeg = 0;  /* reset mbeg to stop further start match */
	    } if (prm) {
		strncpy(prm, src, nsrc-src);
		prm[nsrc - src] = 0;
	    }
	} else if (prm) {
	    /* end of pattern space */
	    strcpy(prm, src);
	}
	
	/* post-processing of param */
	if (prm && mword) {
	    if (mbeg == -1) {
		/* unanchored '$' start, take last word */
		if ((p = my_strrstr(prm, DELIM, strlen(prm) - 1)) != -1) {
		    tmp = prm + p;
		    do
		      *prm++ = *++tmp;
		    while (*tmp);
		}
	    } else if (!pat[1]) {
		/* '$' at end of pattern, take first word */
		strtok(prm, DELIM);
	    } else {
		/* match only if param is single-worded */
		if (my_strstr(prm, DELIM, strlen(prm) - 1) != -1)
		  return 0;
	    }
	    mbeg = mword = 0;  /* reset match flags */
	}
	src = nsrc + strlen(mpat);
	pat = npat;
    }
    strcpy(var[NUMVAR].str, line);
    return 1;
}

/*
 * Search for #actions to trigger on an input line.
 * The line can't be trashed since we want to print it on the screen later.
 * Return 1 if line matched to some #action, 0 otherwise
 */
int search_action(line)
char *line;
{
    actionnode *p;
    params *param;
    int ret = 0;
    
    if (!(param = malloc(sizeof(params)))) {
	errmsg("malloc");
	return -1;
    }
    set_param(param);
    if (error) return 0;
    
    while (*line == '\r') line++;
    for (p = actions; p; p = p->next) {
        if (p->active && 
	    ((p->type == ACTION_WEAK && match_weak_action(p->pattern, line))
#ifdef USE_REGEXP
	     || (p->type == ACTION_REGEXP && match_regexp_action(p->regexp, line))
#endif
	     )) {
	    ret = 1;
	    error = 0;
	    internal = -1;
            parse_instruction(p->command, 0, 1);
	    history_done = 0;
            break;
	}
    }
    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
      restore_param();
    free(param);
    return ret;
}

/*
 * Search for #prompt actions to trigger on a prompt line.
 * The line can't be trashed since we want to print it on the screen later.
 * Return 1 if line matched to some #prompt, 0 otherwise
 */
int search_prompt(line)
char *line;
{
    promptnode *p;
    params *param;
    int ret = 0;
    
    if (!(param = malloc(sizeof(params)))) {
	errmsg("malloc");
	return -1;
    }
    set_param(param);
    if (error) {
	free(param);
	return 0;
    }
    
    while (*line == '\r') line++;
    for (p = prompts; p; p = p->next) {
        if (p->active && 
	    ((p->type == ACTION_WEAK && match_weak_action(p->pattern, line))
#ifdef USE_REGEXP
	     || (p->type == ACTION_REGEXP && match_regexp_action(p->regexp, line))
#endif
	     )) {
	    ret = 1;
	    error = 0;
	    internal = -1;
            parse_instruction(p->command, 0, 1);
	    history_done = 0;
            break;
	}
    }
    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
      restore_param();
    free(param);
    return ret;
}

/*
 * match mark containing & and $ and return 1 if matched, 0 if not
 */
int match_mark(pat, src, start, end)
char *pat, *src, **start, **end;
{
    char mpat[LINELEN], *npat=0, *npat2=0, *nsrc=0, *prm=0, *endprm = NULL, *tmp, c;
    int mbeg = 0, mword = 0, p;
    
    *start = *end = 0;
    
    if (ISMARKWILDCARD(*pat))
	mbeg = - 1;  /* pattern starts with '&' or '$' */

    while (pat && (c = *pat)) {
	if (ISMARKWILDCARD(c)) {
	    /* & matches a string */
	    /* $ matches a single word */
	    prm = src;
	    if (c == '$')
		mword = 1;
	    else if (!*start)
		*start = src;
	    ++pat;
	}
	
	npat  = first_valid(pat, '&');
	npat2 = first_valid(pat, '$');
	if (npat2 < npat) npat = npat2;
	if (!*npat) npat = 0;
	
	if (npat) {
	    strncpy(mpat, pat, npat-pat);
	    mpat[npat - pat] = 0;
	} else
	  strcpy(mpat, pat);
	
	if (*mpat) {
	    nsrc = strstr(src, mpat);
	    if (!nsrc) return 0;
	    endprm = nsrc;
	    if (!*start)
		if (prm)
		    *start = src;
		else
		    *start = nsrc;
	    *end = nsrc + strlen(mpat);
	} else if (prm) {           /* end of pattern space */
	  *end = endprm = prm + strlen(prm);
	} else
	    *end = src;
	
	
	/* post-processing of param */
	if (mword) 
	    if (prm) {
		if (mbeg == -1) {
		    if (!*pat)
			/* the pattern is "$", take first word */
			p = - 1;
		    else 
			/* unanchored '$' start, take last word */
			p = my_strrstr(prm, DELIM, endprm - prm - 1);
		    *start = prm += p + 1;
		} else if (!*pat) {
		    /* '$' at end of pattern, take first word */
		    if ((p = my_strstr(prm, DELIM)) != -1)
			*end = endprm = prm + p;
		} else {
		    /* match only if param is single-worded */
		    for (tmp = prm; tmp < endprm; tmp++)
			if (strchr(DELIM, *tmp))
			    return 0;
		}
		mbeg = mword = 0;  /* reset match flags */
	    } else
		return 0;
	
	src = nsrc + strlen(mpat);
	pat = npat;
    }
    return 1;
}

/*
 * scan line for marked text and add correct marks
 */
void put_marks(dst, line)
char *dst, *line;
{
    marknode *np, *mfirst;
    char *p, *q, *first, *last;
    char begin[CAPLEN], end[CAPLEN];
    char *endline = line + strlen(line);
    int matchlen;
    
    if (*line) do {
	mfirst = 0;
	first = last = endline;
	for (np = markers; *line && np; np = np->next)
	  if (match_mark(np->pattern, line, &p, &q) && p<first) {
	      
	      /*          if ((p = strstr(line, np->pattern)) && p<first) { */
	      
	      first = p;
	      last = q;
	      mfirst = np;
	  }
	
	if (mfirst) {
	    attr_string(mfirst->attrcode, begin, end);
	    
	    strncpy(dst, line, matchlen = first - line);
	    dst += matchlen;
	    line += matchlen;
	    
	    strcpy(dst, begin);
	    dst += strlen(begin);
	    
	    strncpy(dst, line, matchlen = last - first);
	    
	    dst += matchlen;
	    line += matchlen;
	    
	    strcpy(dst, end);
	    dst += strlen(end);
	}
    } while (mfirst);
    
    strcpy(dst, line);
}

/*
 * write string to tty, wrapping to next line if needed.
 */
void wrap_print(s)
char *s;
{
    char *p, *pos, c, follow = 1;
    char buf[LINELEN];
    int l, m;
    enum { NORMAL, ESCAPE, BRACKET } state;
#ifdef BUG_ANSI
    char *esc, fg[CAPLEN], bg[CAPLEN], flag;
    int bold = 0, blink = 0, uline = 0, inv = 0, standout = 0;

    fg[0] = bg[0] = 0;
#endif

    while (*s == '\r') tty_putc(*s++);
    l = printstrlen(s);
    
    while (l >= cols - col0 - 1) {
        p = buf; m = 0; state = NORMAL; pos = 0;
	while (m < cols - col0 - 1 && *s && *s != '\r') {
	    *p++ = c = *s++;
	    switch(state) {
	    case NORMAL:
		if (c == '\033') {
#ifdef BUG_ANSI
		    esc = s - 1;
#endif		    
		    state = ESCAPE;
		}
		else if (c >= ' ')
		    m++, l--;
		break;
	    case ESCAPE:
		state = (c == '[') ? BRACKET : NORMAL;
		break;
	    case BRACKET:
#ifdef BUG_ANSI
		if (!pos) pos = s - 1;
#endif
		if (isalpha(c)) {
#ifdef BUG_ANSI
		    if (strncmp(esc, tty_modenorm, s - esc) == 0)
			standout = uline = blink = bold = fg[0] = bg[0] = 0;
		    else if (strncmp(esc, tty_modebold, s - esc) == 0)
			bold = 1;
		    else if (strncmp(esc, tty_modeblink, s - esc) == 0)
			blink = 1;
		    else if (strncmp(esc, tty_modeuline, s - esc) == 0)
			uline = 1;
		    else if (strncmp(esc, tty_modeinv, s - esc) == 0)
			inv = 1;
		    else if (strncmp(esc, tty_modestandon, s - esc) == 0)
			standout = 1;
		    else if (strncmp(esc, tty_modestandoff, s - esc) == 0)
			standout = 0;
		    else if (c == 'm') {		/* these are attr changes */
			while (pos != s) {
			    if (isdigit(*pos)) {
				char *end = pos;
				while (isdigit(*++end));

				switch(*pos) {
				case '3':	       		/* 3 : FG */
				    strncpy(fg, pos, end - pos);
				    break;
				case '9':			/* 9 : FG */
				    strncpy(fg, pos, end - pos);
				    break;
				case '1':			/* 10: BG */
				    if (*(pos + 1) != '0')
					break;
				case '4':			/* 4 : BG */
					strncpy(bg, pos, end - pos);
				    break;
				}
				pos = end;
			    }
			    ++pos;
			}
		    }
		    pos = 0;
#endif
		    state = NORMAL;
		}
		break;
	    }
	}
	
#ifdef BUG_ANSI
#define CHK(flag, str) { if (flag) p += sprintf(p, (str)); }

	if (fg[0] || bg[0] || blink || bold || uline || inv || standout)
	    p += sprintf(p, tty_modenorm);
	*p++ = '\n';
	if ((flag = fg[0])) p += sprintf(p, "\033[%s", fg);
	if (bg[0]) {
	    p += sprintf(p, "%s%s", flag ? ";" : "\033[", bg);
	    flag = 1;
	}
	if (flag) *p++ = 'm';
	CHK(blink, tty_modeblink);
	CHK(bold, tty_modebold);
	CHK(uline, tty_modeuline);
	CHK(inv, tty_modeinv);
	CHK(standout, tty_modestandon);
	*p = 0;
	tty_puts(buf);
#else
	*p = '\0';
	tty_printf("%s\n", buf);
#endif
	col0 = 0;
	if (!*s)
	    follow = 0;
    }
    if (follow)
	tty_printf("%s\n", s);
}

void smart_print(line)
char *line;
{
    static char buf[LINELEN];
    
    put_marks(buf, line);
    
    if (opt_wrap)
      wrap_print(buf);
    else
      tty_printf("%s\n", buf);
}

/*
 * write string to file, stripping \r
 */
void writetofile(f, str, len)
FILE *f; char *str; int len;
{
    for (; len--; str++) if (*str != '\r') fputc(*str, f);
}

/*
 * insert a typed character on screen (if it is printable)
 */
void insert_char(c)
char c;
{
    if (edlen < LINELEN - 2 && (c >= ' ' || c < 0)) {
	tty_insert_follow_chars(&c, 1);
	if (ISRPAREN(c)) flashparen(c);
    }
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
void get_user_input()
{
    int i, j, chunk = 1;
    static char buf[BUFSIZE+1];   /* allow for terminating \0 */
    char *c = buf;
    static char typed[CAPLEN];    /* chars typed so far (with partial match) */
    static int nchars = 0;	  /* number of them */
    
    /* We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line mode, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    if (!(linemode & (LM_NOECHO | LM_CHAR))) /* line mode, local echo */
	chunk = BUFSIZE;
    
    while ((j = tty_read(tty_read_fd, c, chunk)) < 0 && errno == EINTR)
	;
    if (j <= 0 || (chunk == 1 && j != chunk))
	syserr("read from tty");
    
    c[chunk] = '\0';
    
    if (linemode & LM_CHAR) {
	/* char mode. chunk == 1 */
	while ((i = write(tcp_fd, c, 1)) < 0 && errno == EINTR)
	    ;
	if (i != 1)
	    syserr("write to socket");
	if (!(linemode & LM_NOECHO))
	    tty_putc(*c);
	last_edit_cmd = 0;
    } else if (linemode & LM_NOECHO) {
	/* sending password (line mode, no echo). chunk == 1 */
	if ((*c != '\n' && *c != '\r') && edlen < LINELEN - 2)
	    edbuf[edlen++] = *c;
	else {
	    edbuf[edlen] = '\0';
	    tty_putc('\n');
	    tcp_write(edbuf);
	    edlen = 0;
	    typed[nchars = 0] = 0;
	    col0 = surely_isprompt = '\0';
	    promptstr[0] = promptlen = 0;
	}
	edbuf[pos = edlen] = '\0';
	last_edit_cmd = 0;
    } else {
	/* normal mode (line mode, echo). chunk == BUFSIZE */
	int done = 0;
	keynode *p;
	confirm = 0;
	
	for (; j > 0; c++, j--) {
	    
	    /* search function key strings for match */
	    /* GH: support for \0 in sequence */
	    done = 0;    
	    typed[nchars++] = *c;
	    
	    while (!done) {
		done = 1;
		/*
		 * shortcut:
		 * a single ASCII char at start cannot match any #bind
		 */
		if (nchars == 1 && *c >= ' ' && *c <= '~')
		    p = NULL;
		else {
		    for (p = keydefs; (p && (p->seqlen < nchars || 
					     memcmp(typed, p->sequence, nchars)));
			 p = p->next)
			;
		}
		
		if (!p) {
		    /* 
		     * GH: type the first character and keep processing
		     *     the rest in the input buffer
		     */
		    i = 1;
		    last_edit_cmd = 0;
		    insert_char(typed[0]);
		    while (i < nchars) {
			typed[i - 1] = typed[i];
			i++;
		    }
		    if (--nchars)
			done = 0;
		} else if (p->seqlen == nchars) {
		    p->funct(p->call_data);
		    last_edit_cmd = p->funct;	/* GH: keep track of last command */
		    nchars = 0;
		}
	    }
	}
    }
}


/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
void flashparen(c)
int c;
{
    int lev, i;
    for (i = pos - 1, lev = 0; i >= 0; i--)
      if (ISRPAREN(edbuf[i])) {
	  lev++;
      } else if (ISLPAREN(edbuf[i])) {
	  lev--;
	  if (!lev)
	    if (LPAREN(c) == edbuf[i])
	    break;
	  else
	    i = -1;
      }
    if (i >= 0) {
        tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
        flashback = 1;
        excursion = i;
    }
}

/*
 * put cursor back where it belongs
 */
void putbackcursor()
{
    tty_gotoxy_opt(CURCOL(excursion), CURLINE(excursion), CURCOL(pos), CURLINE(pos));
    flashback = 0;
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
void put_in_history(str)
char *str;
{
    char *p;
    if (hist[curline]) free(hist[curline]);
    if (!(hist[curline] = str_dup(str))) {
	errmsg("malloc");
	return;
    }
    
    if (++curline == MAXHIST) curline = 0;
    
    /* split into words and put into completion list */
    for (p = strtok(str, WORDDELIM); p; 
	 p = strtok((char*)NULL, WORDDELIM)) {
        if (strlen(p) >= MIN_WORD_LEN && 
	    p[0] != '#') /* no commands/short words */
	  put_word(p);
    }
}

/*
 * match and complete a word referring to the word list
 */
void complete_word()
{
  /*
   * GH: rewritten to allow circulating through history with repetitive command
   *     code stolen from cancan 2.6.3a
   *        curr_word:   index into words[]
   *        comp_len     length of current completition
   *        root_len     length of the root word (before the completition)
   *        root         start of the root word
   */	

    static int curr_word, comp_len, root_len;
    char *root, *p;
    int k, n;

    /* find word start */
    if (last_edit_cmd == complete_word && comp_len) {
	k = comp_len;
	tty_moveto(pos - k);
	n = pos - root_len;
    } else {
	for (n = pos; n > 0 && !ISWORDDELIM(edbuf[n - 1]); n--);
	k = 0; curr_word = wordindex; root_len = pos - n;
    }
    root = edbuf + n; comp_len = 0;

    /* k = chars to delete,  n = position of starting word */

    /* scan word list for next match */
    while ((p = words[curr_word = words[curr_word].next].word)) {
	if (!strncasecmp(p, root, root_len) &&
	    *(p += root_len) &&
	    (n = strlen(p)) + edlen < LINELEN) {
	    comp_len = n;
	    while (k && n) {
		k--; n--;
		tty_overtype_follow(*p++);
	    }
	    if (n) tty_insert_follow_chars(p, n);
	    break;
	}
    }
    if (k) tty_delete_nofollow_chars(k);

	/* delete duplicate instances of the word */
    if (p && !(words[k = curr_word].flags & WORD_UNIQUE)) {
	words[k].flags |= WORD_UNIQUE;
	p = words[k].word;  
	n = words[k].next;
	while (words[k = n].word) {
	    n = words[k].next;
	    if (!strcmp(p, words[k].word)) {
		demote_word(k);
		free(words[k].word); words[k].word = 0;
		words[curr_word].flags |= words[k].flags;	/* move retain flag */
		if ((words[k].flags &= WORD_UNIQUE)) break;
	    }
	}
    }
}

/*
 * match and complete entire lines backwards in history
 * GH: made repeated complete_line cycle through history
 */
void complete_line()
{
    static int curr_line, root_len, first_line;
    int i;

    if (last_edit_cmd != complete_line) {
	root_len = edlen;
	first_line = curr_line = curline;
    }

    for (i = curr_line - 1; i != curr_line; i--) {
	if (i < 0) i = MAXHIST - 1;
	if (i == first_line) break;
	if (hist[i] && !strncmp(edbuf, hist[i], root_len)) break;
    }
    if (i != curr_line) {
	tty_clear_input_line(0);
	if (i == first_line) {
	    edbuf[root_len] = 0;
	    edlen = root_len;
	} else {
	    strcpy(edbuf, hist[i]);
	    edlen = strlen(edbuf);
	}
	pos = edlen;
	tty_draw_input_line();
	curr_line = i;
    }
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
    char *p, *q, *r;
    char buf[LINELEN];
    int n;
    
    strcpy(buf, edbuf);
    p = buf; r = edbuf;
    while (*p) 
    {
	while (*p && !isalnum(*p)) *r++ = *p++;
	q = p;
	while (*p && isalnum(*p)) p++;
	if (*p || *q) 
	{
	    n = p - q;
	    if (!strncasecmp(q, "u", n))
	      strcpy(r, "you"); r += 3;
	    else if (!strncasecmp(q, "ur", n))
	      strcpy(r, "your"); r += 4;
	    else if (!strncasecmp(q, "urs", n))
	      strcpy(r, "yours"); r += 5;
	    else if (!strncasecmp(q, "r", n))
	      strcpy(r, "are"); r += 3;
	    else if (!strncasecmp(q, "4", n))
	      strcpy(r, "for"); r += 3;
	    else if (!strncasecmp(q, "2", n))
	      strcpy(r, "to"); r += 2;
	    else if (!strncasecmp(q, "b", n))
	      strcpy(r, "be"); r += 2;
	    else if (!strncasecmp(q, "b4", n))
	      strcpy(r, "before"); r += 6;
	    else
	      strcpy(r, q); r += n;
	}
    }
    *r = '\0';
    edlen = r - edbuf;
}
#endif

/*
 * put string into input line (will be redrawn next time)
 */
void to_input_line(str)
char *str;
{
    strcpy(inserted_next, str);
}

/* editor utility functions: */

/*
 * terminate powwow as cleanly as possible
 */
void exitpowwow()
{
    if (capfile) fclose(capfile);
    if (recordfile) fclose(recordfile);
    tty_puts(edattrend);
    (void)save_settings();
    show_stat();
    tty_quit();
    exit(0);
}


/*
 * GH: Sets a signal-handler permanently (like bsd signal())
 *     from Ilie
 */
void (*sig_permanent(signum, sighandler))()
int signum;
void (*sighandler)();
{
    struct sigaction act;
    void (*oldhandler)();
    if (sigaction(signum, NULL, &act))
        return SIG_ERR;
    oldhandler = act.sa_handler;
    act.sa_handler = sighandler;
    act.sa_flags = 0;
    if (sigaction(signum, &act, NULL)) {
	return SIG_ERR;
    } else {
	return oldhandler;
    }
}

/*
 * set up our signal handlers
 */
void signal_init()
{
    sig_permanent(SIGCHLD, childdeathsig);
    sig_permanent(SIGINT, intrsig);
    sig_permanent(SIGQUIT, intrsig);
    sig_permanent(SIGTERM, termsig);
    sig_permanent(SIGPIPE, pipesig);
    sig_permanent(SIGWINCH, tty_window_change);
    if (sig_permanent(SIGTSTP, SIG_IGN) == SIG_DFL) {
	sig_permanent(SIGTSTP, suspend_powwow);
	can_suspend = 1;
    }
}

void pipesig()
{
    tty_puts("\n#broken pipe.\n");
}

void intrsig()
{
    if (confirm) {
	tty_puts("\n#interrupted.\n");
	exitpowwow();
    }
    
    tty_puts("\n#interrupted. Press again to quit\n");
    if (ready) {
	showprompt();
	tty_draw_input_line();
    }
    tty_flush();
    gettimeofday(&now, NULL);
    confirm = 1;
    error = USER_BREAK;
}

void termsig()
{
    tty_puts("\n#termination signal.\n");
    exitpowwow();
}

/*
 * suspend ourselves
 */
void suspend_powwow()
{
    if (can_suspend) {
        sig_permanent(SIGTSTP, SIG_DFL);
        sig_permanent(SIGTERM, SIG_IGN);
        sig_permanent(SIGINT, SIG_IGN);
        tty_puts(edattrend);
        tty_quit();
	
	if (kill(0, SIGTSTP) < 0) {
	    errmsg("suspend powwow");
	    return;
	}
	
        signal_init();
        tty_init();
        tty_window_change();               /* in case size changed meanwhile */
	/* GH: we shouldn't reset col0 here! */
        tty_gotoxy(0, line0 = lines - 1);
	/* so we know where the cursor is */
        tty_printf("%s%s", edattrend, promptstr);
        tty_draw_input_line();
    } else
      tty_puts("\n#I don't think your shell has job control.\n");
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    if (msg && *msg) {
	tty_clear_input_line(opt_compact);
	if (!opt_compact)
	    tty_putc('\n');
	tty_flush();
	
	fprintf(stderr, "#powwow: fatal system call error:\n\t%s (%d", msg, errno);
	if (errno > 0 && errno < sys_nerr)
	  fprintf(stderr, ": %s", strerror(errno));
	fprintf(stderr, ")\n");
    }
    
#ifdef SAVE_ON_SYSERR
    /* Try to do an emergency save. This can wreak havoc
     * if called from the wrong place, like
     * read_settings() or save_settings(),
     * thus is executed only if you add -DSAVE_ON_SYSERR
     * to CF flags in make_it
     */
    (void)save_settings();
#endif
    
    tty_quit();
    exit(1);
}

void errmsg(msg)
char *msg;
{
    if (!msg)
      msg = "";
    
    tty_clear_input_line(opt_compact);
    if (!opt_compact)
      tty_putc('\n');
    if (errno == EINTR) {
	PRINT_INT("#user break: %s (%d: %s)\n",
		  msg, errno, strerror(errno));
    } else {
	PRINT_INT("#system call error: %s (%d", msg, errno);
	if (errno > 0 && errno < sys_nerr)
	  tty_printf(": %s", strerror(errno));
	tty_printf(")\n");
    }
    tty_flush();
}

void set_param(altparam)
params *altparam;
{
    int i;
    
    if (paramstk.curr < MAX_STACK) {
	paramstk.p[paramstk.curr++] = (params *)var[NUMVAR].num;
        /* it works, because var[NUMVAR].num points to param.num[0],
         * which is the first field of param.
         */
	
	for (i=0; i<NUMPARAM; i++) {
	    *(var[i+NUMVAR].num = &altparam->num[i]) = 0;
	    *(var[i+NUMVAR].str = altparam->str[i]) = 0;
	}
    }
    else {
	print_error(error=DYN_STACK_OV_ERROR);
	paramstk.curr = 0;    /* reset stack to empty */
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &def_param.num[i];
	    var[i+NUMVAR].str = def_param.str[i];
	}
    }
}

void restore_param()
{
    int i;
    params *v;
    
    if (paramstk.curr > 0) {
	v = paramstk.p[--paramstk.curr];
	
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &v->num[i];
	    var[i+NUMVAR].str = v->str[i];
	}
    }
    else {
	print_error(error=DYN_STACK_UND_ERROR);
	paramstk.curr = 0;    /* reset stack to empty */
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &def_param.num[i];
	    var[i+NUMVAR].str = def_param.str[i];
	}
    }
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk (word, silent, maponly)
char *word, silent, maponly;
{
    long strtol();
    char buf[16];
    int n = strlen(word);
    int is_main = (tcp_fd == tcp_main_fd);
    
    if (!is_main && !maponly && !opt_speedwalk)
      return 0;
    if (!n || (n > 1 && !opt_speedwalk && !maponly) ||
	!strchr("neswud", word[n - 1]) ||
	strspn(word, "neswud0123456789") != n)
      return 0;
    
    if (maponly) silent = 1;
    else if (opt_speedwalk) internal = -1;
    buf[1] = '\0';
    while (*word) {
        if (!silent) tty_putc('[');
        if (isdigit(*word)) {
            n = strtol(word, &word, 10);
	    if (!silent)
	      tty_printf("%d", n);
	} else
	  n = 1;
        if (!silent)
	  tty_putc(*word);
        while (n--) {
            *buf = *word;
            if (!maponly) {
	        tcp_write(buf);
	    }
            if (is_main || maponly)
	      map_add_dir(*word);
        }
	if (!silent)
	  tty_puts("] ");
        word++;
    }
    if (!silent)
      tty_putc('\n');
    return !maponly;
}

/*
 * split str into words separated by DELIM, and place in
 * var[1+NUMVAR].str ... var[9+NUMVAR].str -
 * the whole str is put in var[0+NUMVAR].str
 */
char *split_words (str)
char *str;
{
    int i;
    char *prm;
    
    strcpy(var[NUMVAR].str, str);
    for (i=NUMVAR + 1; i < NUMPARAM+NUMVAR; i++) {
	*var[i].num = 0;
	prm = var[i].str;
	while (*str && strchr(DELIM, *str))
	  str++;
	while (*str && !strchr(DELIM, *str))
	  *prm++ = *str++;
	*prm = 0;
    }
    return str;
}

/*
 * copy first word of src into dst, and return pointer to second word of src
 */
char *split_first_word(dst, src)
char *dst, *src;
{
    char *tmp, *delim = DELIM;
    int what, got = 0;
    
    tmp = src = skipspace(src);
    
    if (!*src) {
	*dst='\0';
	return src;
    }
    
    do
      for  (what = 0; delim[what] && !got; what++)
      if (*tmp == delim[what])
      got = 1;
    while (!got && *++tmp);
    
    strncpy(dst, src, tmp-src);
    dst[tmp-src] = '\0';
    
    if (*tmp && *tmp != CMDSEP) tmp++;
    return tmp;
}

/*
 * find the first valid (non-escaped)
 * char in a string
 */
char *first_valid(p, ch)
char *p, ch;
{
    int escaped = 0;
    
    if (ch == ESC2)
	while (*p && *p != ESC2)
	    p++;
    else while (*p && ((*p != ch) || escaped)) {
	if (*p == ESC)
	    escaped = 1;
	else
	    escaped = 0;
	p++;
    }
    return p;
}

/*
 * find the first regular (non-escaped, non in "" () or {} )
 * char in a string
 */
char *first_regular(p, c)
char *p, c;
{
    int escaped = 0, quotes=0, paren=0, braces=0;

    /* GH: p = skipspace(p); */
    
    while (*p && ((*p != c) || escaped || quotes || paren>0 || braces>0)) {
	if (*p == ESC2) {
	    escaped = 0;
	    while (*p == ESC2)
		p++;
	}
	else if (*p == ESC) {
	    escaped = 1;
	    while (*p == ESC)
		p++;
	}
	else {
	    escaped = 0;
	    if (quotes) {
		if (*p == '\"')
		    quotes = 0;
	    }
	    else if (*p == '\"')
		quotes = 1;
	    else if (*p == ')')
		paren--;
	    else if (*p == '(')
		paren++;
	    else if (*p == '}')
		braces--;
	    else if (*p == '{')
		braces++;
	    p++;
	}
    }
    return p;
}

char *get_next_instr(p)
char *p;
{
    int count, is_if;
    char *sep, *q;
    
    if (!*p)
	return p;
    
    
    p = skipspace(p);
    
    count = is_if = !strncmp(p, "#if", 3);
    
    do {
	sep   = first_regular(p, CMDSEP);
	
	q = p;
	if (*q) do {
	    if (*q == '#') q++;
	    
	    q = first_regular(q, '#');
	} while (*q && strncmp(q, "#if", 3));
	
	if (sep<=q) {
	    if (*(p = sep))
		p++;
	}
	else
	    if (*q)
		p = get_next_instr(q);
	else {
	    print_error(error=SYNTAX_ERROR);
	    return NULL;
	}
	sep = skipspace(p);    
    } while (*p && count-- && 
	     (!is_if || (!strncmp(sep, "#else", 5) &&
			 (*(p = sep + 5)))));
    
    return p;
}

/*
 * Parse and exec the first instruction in 'line', and return pointer to the
 * second instruction in 'line' (if any).
 */
char *parse_instruction(line, silent, subs)
char *line; char silent, subs;
{
    aliasnode *np;
    char buf[LINELEN], substbuf[LINELEN];
    char *arg, *end, *ret;
    char last_is_sep = 0;
    int otcp_fd = -1;
    
    if (error) return NULL;
    
    ret = get_next_instr(line);
    
    if (!ret || ret==line)  /* error or empty instruction, bail out */
      return ret;

    /*
     * remove the optional ';' after an instruction,
     * to have an usable string, ending with \0.
     * If it is escaped, DON'T remove it: it is not a separator,
     * and the instruction must already end with \0, or we would not be here.
     */
    if (ret[-1] == CMDSEP) {
	/* instruction is not empty, ret[-1] is allowed */
	if (ret > line + 1 && ret[-2] == ESC) {
	    /* ';' is escaped */
	} else {
	    *--ret = '\0';
	    last_is_sep = 1;
	}
    }

    if (subs)
      subst_param(substbuf, line);
    else
      strcpy(substbuf, line);

    /* restore integrity of original line: must still be put in history */
    if (last_is_sep)
	*ret++ = CMDSEP;

    /* run-time debugging */
    if (opt_debug) tty_printf("#parsing: %s\n", substbuf);
    
    arg = skipspace(line = substbuf);
    
    if (arg[0] == '#' && arg[1] == '#') { /* send to other connection */
	line = split_first_word(buf, arg + 2);
	otcp_fd = tcp_fd;
	if ((tcp_fd = tcp_find(buf))<0) {
	    tty_printf ("#no connection named `%s'\n", buf);
	    tcp_fd = otcp_fd;
	    return NULL;
	}
	arg = skipspace(line);
	if (!*arg) {
	    if (CONN_LIST(tcp_fd).mode == 0) {
		/* set it as main connection */
		tcp_main_fd = tcp_fd;
		linemode = 0;
		*promptstr = promptlen = 0;
		return ret; /* stop parsing, otherwise a newline would be sent to tcp_fd */
	    } else {
		PRINT_INT ("#only MUD connections can be default ones!\n");
		return NULL;
	    }
	}
    }
    
    if (*arg == '{') {    /* instruction contains a block */
	end = first_regular(line = arg + 1, '}');
	
	if (*end) {
	    *end = '\0';
	    parse_user_input(line, silent);
	    *end = '}';
	}
	else
	  print_error(error=MISSING_PAREN_ERROR);
    
    } else {
	int oneword;
	/* initial spaces are NOT skipped this time */
	
	arg = split_first_word(buf, line); /* buf contains the first word,
					    arg points to arguments */
	if (!*arg) oneword = 1;
	else oneword = 0;
	
	if ((np = *lookup_alias(buf))) {
	    params *param;
	    if (!(param = malloc(sizeof(params)))) {
		errmsg("malloc");
		return ret;
	    }
	    set_param(param);
	    if (error) return NULL;
	    
	    split_words(arg);  /* split argument into words
                                and place them in $0 ... $9 */
	    parse_instruction(np->subst, 0, 1);
	    
	    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
	      restore_param();
	    
	    free(param);
	    
	    /* now check for internal commands */
	    /* placed here to allow also aliases starting with `#' */
	} else if (*(end = skipspace(line)) == '#') {
	    
	    if (*(end = skipspace(end + 1)) == '(') {    /* execute #() */
		end++;
		(void)eval(NULL, &end, PRINT_NOTHING);
	    } else
		parse_commands(buf + 1, arg);
	    /* ok, buf contains skipspace(first word) */
	} else if (!oneword || !parse_walk(buf, silent, 0)) {
	    /* it is ok, parse_walk accepts only one word */
	    
	    unescape(line);
	    if (!silent && echo_ext) tty_printf("[%s]\n", line);
	    col0 = promptlen = promptstr[0] = surely_isprompt = '\0'; 
	    tcp_write(line);
	}
    }
    if (otcp_fd != -1) {
      tcp_fd = otcp_fd;
    }
    
    return ret;
}

/*
 * parse input from user: calls parse_instruction for each instruction
 * in cmd_line.
 * silent = 1 if the line should not be echoed, 0 otherwise.
 */
void parse_user_input(line, silent)
char *line; char silent;
{
    do {
	line = parse_instruction(line, silent, 0);
    } while (!error && line && *line);
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *p;
    p = s;
    while (*s) {
	if (*s == ESC) {
	    s++;
	    if (ISODIGIT(*s) &&
		ISODIGIT(*(s + 1)) &&
		ISODIGIT(*(s + 2))) {
		*p++ = (((*s - '0') << 6) |
			((*(s + 1) - '0') << 3) |
			(*(s + 2) - '0'));
		s += 3;
		continue;
	    } else {
		while (*s == ESC)
		    *p++ = *s++;
		if (*s == ESC2) {
		    *p++ = ESC;
		    s++;
		    continue;
		}
	    }
	}
	if (*s != ESC)
	  *p++ = *s++;
    }
    *p = '\0';
}

/*
 * parse powwow's own commands
 */
void parse_commands(command, arg)
char *command, *arg;
{
    int i, j;
    
    /* We ALLOW special commands also on subsidiary connections ! */
    
    /* assume output will be enough to make input line = last screen line */
    line0 = lines - 1;
    if (isdigit(*command) && (i = atoi(command))) {
	if (i > 0)
	    while (i--) (void)parse_instruction(arg, 1, 0);
	else
	    tty_puts("#bogus repeat count\n");
    } else {
	j = strlen(command);
	for (i = 0; commands[i].name; i++)
	  if (!strncmp(command, commands[i].name, j)) {
	      if (commands[i].funct) {
		  tty_puts(edattrend);
		  (*commands[i].funct)(arg);
		  return;
	      }
	  }
	tty_printf("#unknown powwow command `%s'\n", command);
    }
}

/*
 * redisplay the prompt
 */
void showprompt()
{
    char buf[LINELEN];
    put_marks(buf, promptstr);
    tty_printf("%s%s%s", edattrend, buf, edattrbeg);
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will remove the session from the list)
 */
void cancel_edit(sp)
editsess *sp;
{
    char buf[LINELEN];
    char keystr[16];
    
    if (kill(sp->pid, SIGKILL) < 0) {       /* Editicide */
	errmsg("kill editor child");
	return;
    }
    tty_printf("#killed `%s' (%d)\n", sp->descr, sp->key);
    sprintf(keystr, "C%d\n", sp->key);
    sprintf(buf, "%sE%d\n%s", MPI, (int) strlen(keystr), keystr);
    tcp_main_write(buf);
    sp->key = -1;        /* mark the child as cancelled */
}

/*
 * substitute $0..$9 and @0..@9 in a string 
 * (unless $ or @ is escaped with backslash)
 */
void subst_param(dst, src)
char *src, *dst;
{
    int i, subst;
    char *tmp, c;
    
    while (*src)  {
	subst = 0;
	if (*src == '$' || *src == '@') {
	    c = *src;
	    tmp = src + 1;
	    if (isdigit(*tmp)) {
		i = 0;
		while (isdigit(*tmp) && i < NUMPARAM) {
		    i *= 10;
		    i += *tmp++ - '0';
		}
		if (i < NUMPARAM) {
		    src = tmp;
		    subst = 1;
		    if (c == '$') {
			strcpy(dst, var[i+NUMVAR].str);
			if (*dst) while (*++dst) ;
		    } else {
			sprintf(dst, "%ld", *var[i+NUMVAR].num);
			while (*++dst) ;
		    }
		}
	    }
	} 
	if (!subst) {
	    if (*src == ESC) {
		src++;
		while (*src == ESC)
		    *dst++ = *src++;
		
		if (*src)
		    *dst++ = *src++;
	    }
	    else if (*src)
		*dst++ = *src++;
	}
    }
    *dst = '\0';
}

/*
 * return pointer to next non-blank char
 */
char *skipspace(p)
char *p;
{
    while (*p == ' ') p++;
    return p;
}

void load_missing_stuff(n)
int n;
{
    if (n < 1) {
        tty_puts("#default keypad settings loaded\n");
        tty_add_walk_binds();
    }
    if (n < 2) {
        tty_puts("#default editing keys settings loaded\n");
        tty_add_initial_binds();
    }
    if (n < 5) {
	char buf[LINELEN], *str[] = { "compact", "debug", "echo", "info",
		"keyecho", "speedwalk", "wrap", 0 };
	int i;
	tty_puts("#default #option aliases loaded\n");
	for (i=0; str[i]; i++) {
	    sprintf(buf, "#%s={#if ($(1)==\"on\") #option +%s; #else #if ($(1)==\"off\") #option -%s; #else #option %s}",
		    str[i], str[i], str[i], str[i]);
	    parse_alias(buf);
	}
    }
}

/*
 * read definitions from file
 * return > 0 if success, < 0 if fail.
 * NEVER call syserr() from here or powwow will
 * try to save the definition file even if it got
 * a broken or empty one.
 */
int read_settings()
{
    char buf[LINELEN];
    FILE *f;
    char *p, *cmd, done, old_nice = a_nice;
    int failed = 1, n;
    varnode **first;
    
    if (!*deffile) {
	tty_puts("#warning: no save-file defined!\n");
	return -1;
    }
    
    f = fopen(deffile, "r");
    if (!f) {
	PRINT_INT ("#cannot find save-file: %s\n", deffile);
	return -1;
    }
    
    n = echo_int = a_nice = 0;
    
    while (aliases)       delete_aliasnode(&aliases);
    while (actions)       delete_actionnode(&actions);
    while (markers)       delete_marknode(&markers);
    while (keydefs)       delete_keynode(&keydefs);
    while (named_vars[0]) delete_varnode(&named_vars[0], (char)0);
    first = &named_vars[1];
    while (*first) {
	if (is_permanent_variable(*first))
	    first = &(*first)->next;
	else	
	    delete_varnode(first, (char)1);
    }
    
    for (n = 0; n < NUMVAR; n++) {
	*var[n].num = 0;
	if ((p = var[n].str)) {
	    free(p);
	    var[n].str = NULL;
	}
    }
    
    while (fgets(buf, LINELEN, f)) {
	cmd = strtok(buf, "\n");        /* strip lf */
	if (!cmd)
	    continue;
        cmd = skipspace(cmd);
	if (!*cmd)
	    continue;
	
        error = 0;
	done = 0;
        if (*(p = first_regular(cmd, ' '))) {
	    done = 1;
            *p++ = '\0';
            if (!strcmp(cmd, "#savefile-version")) {
                n = atoi(p);
                if (n > SAVEFILEVER)
		    tty_puts("#warning: this powwow version is too old!\n");
                else if (n < SAVEFILEVER) {
		    tty_puts("#warning: config file is from an older version\n");
		    load_missing_stuff(n);
		}
            } else {
	        *--p = ' ';
		done = 0;
	    }
	}
	if (!done)
	    parse_user_input(cmd, 1);
    }
    
    if (ferror(f) && !feof(f)) {
        tty_printf("#error: can't read file `%s': %s\n", deffile, strerror(errno));
        n = 0; failed = -1;
    }
    
    fclose(f);
    a_nice = old_nice;
    
    return failed;
}

static char tmpname[LINELEN];

void fail_msg()
{
    tty_printf("#error: can't write to temporary file `%s': %s\n", tmpname, strerror(errno));
}

/*
 * save settings in definition file
 * return > 0 if success, < 0 if fail.
 * NEVER call syserr() from here or powwow will
 * enter an infinite loop!
 */
int save_settings()
{
    FILE *f;
    int l;
    aliasnode *alp;
    actionnode *acp;
    promptnode *ptp;
    marknode *mp;
    keynode *kp;
    varnode *vp;
    extern char *full_options_string;
    
    char buf[LINELEN], *buf2;
    int i, flag, failed = 1;
    
    if (!*deffile) {
	PRINT_INT("#warning: no save-file defined!\n");
	return -1;
    }
    
    /*
     * Create a temporary file in the same directory as deffile,
     * and write settings there
     */
    strcpy(tmpname, deffile);
    l = strlen(tmpname) - 1;
    while (l && tmpname[l] != '/')
	l--;
    if (l)
	l++;
    
    sprintf(tmpname + l, "tmpsav%d%d", getpid(), rand() >> 8);
    if (!(f = fopen(tmpname, "w"))) {
        fail_msg();
        return -1;
    }
    
    failed = fprintf(f, "#savefile-version %d\n", SAVEFILEVER);
    if (failed > 0 && *hostname)
	failed = fprintf(f, "#host %s %s\n", hostname, portnumber);
    
    if (failed > 0)
	failed = fprintf(f, "#delim %s%s\n",
			 delim_mode_names[word_delim_mode],
			 word_delim_mode == DELIM_CUSTOM ? WORDDELIM : "");
    
    /* main host */
    if (failed > 0 && *initstr)
	failed = fprintf(f, "#init =%s\n", initstr);
    
    for (alp = aliases; alp && failed > 0; alp = alp->next) {
	escape_specials(alp->name, buf2=buf);
	failed = fprintf(f, "#alias %s=%s\n", buf, alp->subst);
    }
    
    for (acp = actions; acp && failed > 0; acp = acp->next) {
        escape_specials(acp->label, buf2=buf);
        while (*buf2) buf2++;
        *buf2++ = ' ';
        escape_specials(acp->pattern, buf2);
        while (*buf2) buf2++;
	failed = fprintf(f, "#action %c%c%s=%s\n",
			 action_chars[acp->type],
			 acp->active ? '+' : '-',
			 buf, acp->command);
    }

    for (ptp = prompts; ptp && failed > 0; ptp = ptp->next) {
        escape_specials(ptp->label, buf2=buf);
        while (*buf2) buf2++;
        *buf2++ = ' ';
        escape_specials(ptp->pattern, buf2);
        while (*buf2) buf2++;
	failed = fprintf(f, "#prompt %c%c%s=%s\n",
			 action_chars[ptp->type],
			 ptp->active ? '+' : '-',
			 buf, ptp->command);
    }
    
    for (mp = markers; mp && failed > 0; mp = mp->next) {
        escape_specials(mp->pattern, buf);
	failed = fprintf(f, "#mark %s=%s\n", buf, attr_name(mp->attrcode));
    }
    /* save value of global variables */
    
    for (flag = 0, i=0; i<NUMVAR && failed > 0; i++) {
        if (var[i].num && *var[i].num) {    /* first check was missing!!! */
	    failed = fprintf(f, "%s@%d = %ld", flag ? ", " : "#(",
			     i-NUMVAR, *var[i].num);
            flag = 1;
        }
    }
    if (failed > 0 && flag) failed = fprintf(f, ")\n");
    
    for (i=0; i<NUMVAR && failed > 0; i++) {
        if (var[i].str && *var[i].str) {    /* first check was missing!!! */
	    escape_specials(var[i].str, buf);
	    failed = fprintf(f, "#($%d = \"%s\")\n", i-NUMVAR, buf);
	}
    }
    
    for (flag = 0, vp = named_vars[0]; vp && failed > 0; vp = vp->next) {
        if (vp->num) {
	    failed = fprintf(f, "%s@%s = %ld", flag ? ", " : "#(",
			     vp->name, vp->num);
            flag = 1;
	}
    }
    if (failed > 0 && flag) failed = fprintf(f, ")\n");
    
    for (vp = named_vars[1]; vp && failed > 0; vp = vp->next) {
        if (!is_permanent_variable(vp) && *vp->str) {
	    escape_specials(vp->str, buf);
	    failed = fprintf(f, "#($%s = \"%s\")\n", vp->name, buf);
	}
    }
    
    /* GH: fixed the history and word completions saves */
    if (opt_history) {
	for (l = curline - 1; l != curline && failed > 0; l--) {
	    if (l < 0) l += MAXHIST;
	    if (!hist[l])
		break;
	    else {
		escape_specials(hist[l], buf);
		failed = fprintf(f, "#put %s\n", buf);
	    }
	}
    }
    
    if (opt_words) {
	int cl = 4, len;
	l = wordindex;
	flag = 0;
	while (words[l = words[l].next].word);
	while (words[l = words[l].prev].word && failed > 0) {
	    if (~words[l].flags & WORD_RETAIN) {
		escape_specials(words[l].word, buf);
		len = strlen(buf) + 1;
		if (cl + len >= 80 - 4) {
		    cl = 4;
		    failed = fprintf(f, "\n");
		    flag = 0;
		}
		failed = fprintf(f, "%s %s", flag ? "" : "#add", buf);
		cl += len;
		flag = 1;
	    }
	}
	if (failed > 0 && flag)
	    failed = fprintf(f, "\n");
    }
    
    for (kp = keydefs; kp && failed > 0; kp = kp->next) {
	/* 	escape_specials(seq_name(kp->sequence, kp->seqlen), buf); */
	failed = fprintf(f, "#bind %s %s=%s\n", kp->name, 
			 seq_name(kp->sequence, kp->seqlen),
			 /* buf, */
			 kp->funct==key_run_command ? kp->call_data :
			 internal_functions[lookup_edit_function(kp->funct)].name);
    }
    
    if (failed > 0)
	failed = fprintf(f, full_options_string,
			 opt_exit  ? '+' : '-',
			 opt_history ? '+' : '-',
			 opt_words ? '+' : '-',
			 opt_compact ? '+' : '-',
			 opt_debug ? '+' : '-',
			 echo_ext  ? '+' : '-',
			 echo_int  ? '+' : '-',
			 echo_key  ? '+' : '-',
			 opt_speedwalk ? '+' : '-',
			 opt_wrap ? '+' : '-',
			 opt_autoprint  ? '+' : '-', "\n");
    fclose(f);
    
    if (failed <= 0) {
	failed = -1;
	fail_msg();
    }   
    else {
	failed = rename(tmpname, deffile);
	if (failed == -1) {
	    PRINT_INT ("#error: can't move temporary file `%s' to `%s': %s\n",
		       tmpname, deffile, strerror(errno));
	} else {
	    failed = 1;
	}
    }
    
    return failed;
}

/*
 * GH: word history handling stolen from cancan 2.6.3a
 */

void default_completions()
{
    char buf[LINELEN];
    cmdstruct *p;
    int i;
    for (i = 0, buf[0] = '#', p = commands; p->name; p++)
	if (p->funct /*&& strlen(p->name) >= 3*/ ) {
	    if (++i >= MAXWORDS) break;
	    strcpy(buf + 1, p->name);
	    if (!(words[i].word = str_dup(buf)))
	      syserr("malloc");
	    words[i].flags = WORD_UNIQUE | WORD_RETAIN;
	}
    for (i = MAXWORDS; i--; words[i].prev = i - 1, words[i].next = i + 1);
    words[0].prev = MAXWORDS - 1;  words[MAXWORDS - 1].next = 0;
}

/*
 * move a node before wordindex, i.e. make it the last word
 */
static void demote_word(i)
int i;
{
    words[words[i].prev].next = words[i].next;
    words[words[i].next].prev = words[i].prev;
    words[i].prev = words[words[i].next = wordindex].prev;
    words[wordindex].prev = words[words[wordindex].prev].next = i;
}

/*
 * put word in word completion ring
 */
void put_word(s)
char *s;
{
    int r = wordindex;
    if (!(words[r].word = str_dup(s))) {
	errmsg("malloc");
	return;
    }
    words[r].flags = 0;
    while (words[r = words[r].prev].flags & WORD_RETAIN)
      ;
    demote_word(r);
    wordindex = r;
    if (words[r].word) { 
	free(words[r].word); 
	words[r].word = 0; 
    }
}

/*
 * GH: set delimeters[DELIM_CUSTOM]
 */
void set_custom_delimeters(s)
char *s;
{
    if (delimeters[DELIM_CUSTOM])
      free(delimeters[DELIM_CUSTOM]);
    if (!(delimeters[DELIM_CUSTOM] = str_dup(s)))
      errmsg("malloc");
}

/*
 * GH: return true if var is one of the permanent variables
 */
int is_permanent_variable(var)
varnode *var;
{
    return (var == variable_prompt ||
	    var == variable_line);
}
